# -*- coding: utf-8 mode: python -*- vim:sw=4:sts=4:et:ai:si:sta:fenc=utf-8

__all__ = (
    'load',
    'About',
    'Alert', 'set_alert',
    'Menu', 'Action',
)

import sys, traceback

from ..base import seqof, Undef
from ..ext import web

from . import ui
from .ui import Action, u, p
from .api import _ensure_session

Undef = Undef() # spécifique à ce module

def load(prefix=None, icons=True, charset="utf-8"):
    if prefix is None: prefix = "/static/bootstrap/"
    elif not prefix.endswith("/"): prefix = prefix + "/"
    lines = [
        '<meta charset="%s" />' % charset,
        '<meta http-equiv="X-UA-Compatible" content="IE=edge" />',
        '<meta name="viewport" content="width=device-width, initial-scale=1.0" />',
    ]
    if icons:
        lines.extend([
            '<link rel="shortcut" href="%sfavicon.ico" />' % prefix,
            '<link rel="icon apple-touch-icon" href="%sicon.png" />' % prefix,
        ])
    lines.extend([
        ui.css("%sbootstrap/css/bootstrap.min.css" % prefix),
        ui.css("%sbslocal.css" % prefix),
        ui.js("%sjquery.min.js" % prefix),
        ui.js("%sbootstrap/js/bootstrap.min.js" % prefix),
        u"<!--[if lt IE 9]>",
        ui.js("%shtml5shiv/html5shiv.js" % prefix),
        ui.js("%srespond/respond.min.js" % prefix),
        u"<![endif]-->",
    ])
    return u"\n".join(lines)

def sa(value, kw, name, default=Undef):
    """si value est indéfini, récupérer la valeur avec le nom court dans kw
    """
    if value is Undef and name is not None: value = kw.pop(name, Undef)
    if value is Undef: value = default
    return value

################################################################################
# Menu

NAVBAR_TYPE_MAP = {
    'fixed': 'fixed-top',
    'static': 'static-top',
}

class Menu(ui.Menu):
    """Un menu"""

    def __init__(self, title, mitems=None, profiles=None,
                 id=None, in_profiles=None, css=None,
                 navbar_type=None, container_fluid=None, **kw):
        if navbar_type is None: navbar_type = kw.pop('t', None)
        if container_fluid is None: container_fluid = kw.pop('f', None)
        super(Menu, self).__init__(title, mitems, profiles, id, in_profiles, css, **kw)
        if navbar_type is None: navbar_type = 'static'
        self.navbar_type = navbar_type
        if container_fluid is None: container_fluid = True
        self.container_fluid = container_fluid

    def __mitem(self, mitem, sel_profile):
        # ne pas afficher un menu qui n'est pas dans le bon profil
        if sel_profile is None:
            if mitem.in_profiles is not None:
                return []
        elif mitem.in_profiles is not None:
            if sel_profile not in mitem.in_profiles:
                return []
        active = ' class="active"' if mitem.active else ''
        url = web.websafe(mitem.url)
        css = ui.classif(mitem.css)
        accesskey = ui.accesskeyif(mitem.accesskey)
        title = web.websafe(mitem.title)
        return [u"""<li%s><a href="%s"%s%s>%s</a></li>""" % (active, url, css, accesskey, title)]

    def __dropdown(self, menu, profile):
        children = []
        for mitem in menu.mitems:
            children.extend(self.__mitem(mitem, profile))
        if not children:
            # ne pas afficher un menu vide
            return []
        active = ' active' if menu.active else ''
        lines = [u"""<li class="dropdown%s">""" % active]
        firsturl = web.websafe(menu.mitems[0].url)
        title = web.websafe(menu.title)
        lines.append(u"""<a href="%s" class="dropdown-toggle" data-toggle="dropdown">%s <b class="caret"></b></a>""" % (firsturl, title))
        lines.append(u"""<ul class="dropdown-menu">""")
        lines.extend(children)
        lines.append(u"""</ul>""")
        return lines

    def __is_navbar_fixed(self, navbar_type):
        return navbar_type == 'fixed' or navbar_type == 'fixed-top'
    def __is_navbar_static(self, navbar_type):
        return navbar_type == 'static' or navbar_type == 'static-top'

    def render(self, select=None, profile=None, navbar_type=None, container_fluid=None, **kw):
        if select is None: select = kw.pop('s', None)
        if profile is None: profile = kw.pop('p', None)
        if navbar_type is None: navbar_type = kw.pop('t', None)
        if container_fluid is None: container_fluid = kw.pop('f', None)

        if navbar_type is None: navbar_type = self.navbar_type
        if container_fluid is None: container_fluid = self.container_fluid
        if select is not None: self.select(select, profile)
        selection = self.get_mitem()
        lines = []
        if self.__is_navbar_fixed(navbar_type):
            lines.append(u"""<style type="text/css">body { margin-top: 50px; }</style>""")
        css = u''
        css = ui.addclassif('navbar-fixed-top', self.__is_navbar_fixed(navbar_type), css)
        css = ui.addclassif('navbar-static-top', self.__is_navbar_static(navbar_type), css)
        if self.profiles is not None and self.sel_profile is not None:
            css = ui.addclassif('%s-profile' % self.sel_profile, None, css)
        css = ui.addclassif(self.css, None, css)
        lines.append(u"""<div class="navbar navbar-default%s" role="navigation">""" % css)
        lines.append(u"""<div class="container%s">
<div class="navbar-header">
<button type="button" class="navbar-toggle" data-toggle="collapse" data-target=".navbar-collapse">
<span class="sr-only">Tooggle navigation</span>
<span class="icon-bar"></span>
<span class="icon-bar"></span>
<span class="icon-bar"></span>
</button>
</div>
<div class="navbar-collapse collapse">
<ul class="nav navbar-nav">""" % ('-fluid' if container_fluid else ''))
        if self.profiles is not None:
            # gestion des profils
            lines.append(u"""<li class="navbar-text"><form class="form-inline" method="get">""")
            for profile in self.profiles:
                if selection is not None:
                    if profile == self.sel_profile \
                       and selection.in_profiles is not None \
                       and profile in selection.in_profiles:
                        lines.append(u"""<button class="btn btn-xs btn-primary active" type="submit" formaction="%s" formmethod="%s"><span>%s</span></button>""" % (selection.url, selection.method, profile))
                        continue
                    profile_ref = '%s_ref' % profile
                    profile_mitem = None
                    if profile_ref in selection:
                        profile_mitem = self.get_mitem(selection[profile_ref])
                    if profile_mitem is not None:
                        lines.append(u"""<button class="btn btn-xs btn-default" type="submit" formaction="%s" formmethod="%s"><span>%s</span></button>""" % (profile_mitem.url, profile_mitem.method, profile))
                        continue
                lines.append(u"""<span class="btn btn-xs btn-default disabled">%s</span>""" % profile)
            lines.append(u"""</form></li>""")
        for mitem in self.mitems:
            if isinstance(mitem, ui.Menu):
                lines.extend(self.__dropdown(mitem, self.sel_profile))
            else:
                lines.extend(self.__mitem(mitem, self.sel_profile))
        lines.append(u"""</ul>""")
        title = self.title
        if title is not None:
            title = web.websafe(title)
            lines.append(u"""<p class="navbar-text"><b><em>%s</em></b></p>""" % title)
        lines.append(u"""</div>
</div>
</div>""")
        return u"\n".join(lines)
    def __unicode__(self): return self.render()

def set_menu(menu, **params):
    """Décorateur qui permet d'initialiser une instance de menu dans la session
    """
    def decorator(method):
        def wrapper(self, *args, **kw):
            session = _ensure_session(self)
            session.menu = menu
            return method(self, *args, **kw)
        return wrapper
    return decorator

def menu(id=None, profile=None, **params):
    """Décorateur qui permet de mettre à jour la sélection et le profil dans
    le menu

    par défaut, prendre le profil courant dans la session. le profil courant
    est initialisé par le décorateur @set_profile
    """
    if profile is None: profile = params.pop('p', None)
    def decorator(method):
        def wrapper(self, *args, **kw):
            session = _ensure_session(self)
            if 'menu' not in session:
                raise ValueError("menu est requis")
            sel_profile = profile
            if sel_profile is None and 'profile' in session:
                sel_profile = session.profile
            if id is not None:
                session.menu.select(id, sel_profile)
            return method(self, *args, **kw)
        return wrapper
    return decorator

################################################################################
# about

def About(summary, details=None, onShowDetails=None, js=True):
    lines = []
    if js:
        lines.append(u"""<script type="text/javascript">
jQuery.noConflict()(function($) {
  $(".about .show-details").click(function() {
    var $about = $(this).parents("div.about");
    $about.children(".summary").addClass("hide");
    $about.children(".details").removeClass("hide");""")
        if onShowDetails is not None:
            if callable(onShowDetails): onShowDetails = onShowDetails()
            lines.append(onShowDetails)
        lines.append("""\
    return false;
  });
});
</script>""")
    if callable(summary): summary = summary()
    lines.extend([
        u'<div class="about">',
        u'<div class="summary"><p>',
        summary,
    ])
    if details is not None:
        lines.extend([u'<a class="show-details" href="#"><em>En savoir plus...</em></a>'])
    lines.extend([u'</p></div>'])
    if details is not None:
        if callable(details): details = details()
        lines.extend([
            u'<div class="details hide">',
            details,
            u'</div>',
        ])
    lines.extend([u'</div>'])
    return u'\n'.join([u(line) for line in lines])

################################################################################
# Alert

ALERT_TYPE_MAP = dict(
    done="success",
    warn="warning", notice="warning",
    error="danger",
)
class Alert(object):
    msg = None
    exc_info = None
    type = None
    closeable = None
    escape = None
    action = None
    showtb = None

    def __init__(self, msg=None, exc_info=Undef, type=Undef, closeable=Undef, escape=Undef, action=None, showtb=True, **kw):
        exc_info = sa(exc_info, kw, 'e')
        if exc_info is Undef: exc_info = sys.exc_info()
        if exc_info == (None, None, None): exc_info = None
        type = sa(type, kw, 't', "error")
        closeable = sa(closeable, kw, 'c', False)
        escape = sa(escape, kw, 'x', None)

        self.msg = msg
        self.exc_info = exc_info
        self.type = ALERT_TYPE_MAP.get(type, type)
        self.closeable = closeable
        self.escape = escape
        self.action = action
        self.showtb = showtb

    def __call__(self, msg=Undef, exc_info=Undef, type=Undef, closeable=Undef, escape=Undef, action=Undef, showtb=Undef, **kw):
        exc_info = sa(exc_info, kw, 'e')
        if msg is Undef:
            # si on ne spécifie pas de message, alors prendre la valeur actuelle
            msg = self.msg
            if exc_info is Undef: exc_info = self.exc_info
        else:
            # si on spécifie un message, alors prendre aussi l'exception courante
            if exc_info is Undef: exc_info = sys.exc_info()
            if exc_info == (None, None, None): exc_info = None
        type = sa(type, kw, 't', self.type)
        closeable = sa(closeable, kw, 'c', self.closeable)
        escape = sa(escape, kw, 'x', self.escape)
        action = sa(action, kw, None, self.action)
        showtb = sa(showtb, kw, None, self.showtb)

        self.msg = msg
        self.exc_info = exc_info
        self.type = ALERT_TYPE_MAP.get(type, type)
        self.closeable = closeable
        self.escape = escape
        self.action = action
        self.showtb = showtb
        return self

    def render(self, msg=Undef, exc_info=Undef, type=Undef, closeable=Undef, escape=Undef, action=Undef, showtb=Undef, **kw):
        exc_info = sa(exc_info, kw, 'e')
        if msg is Undef:
            # si on ne spécifie pas de message, alors prendre la valeur initiale
            msg = self.msg
            if exc_info is Undef: exc_info = self.exc_info
        else:
            # si on spécifie un message, alors prendre aussi l'exception courante
            if exc_info is Undef: exc_info = sys.exc_info()
        type = sa(type, kw, 't', self.type)
        closeable = sa(closeable, kw, 'c', self.closeable)
        escape = sa(escape, kw, 'x', self.escape)
        action = sa(action, kw, None, self.action)
        showtb = sa(showtb, kw, None, self.showtb)

        if callable(msg):
            # si msg est callable, par défaut ne pas mettre le résultat en
            # échappement
            if escape is None: escape = False
            msg = msg()
        if escape is None: escape = True

        if msg is None and exc_info is not None:
            msg = u"Une erreur inattendue s'est produite"
        if msg is None: return u""
        if escape: msg = web.websafe(msg)

        lines = []
        css = u"alert alert-%s" % ALERT_TYPE_MAP.get(type, type)
        css = ui.addclassif("alert-dismissible", closeable, css)
        lines.append(u"""<div class="%s" role="alert">""" % css)
        if closeable:
            lines.append(u"""<button type="button" class="close" data-dismiss="alert" aria-label="Close"><span aria-hidden="true">&times;</span></button>""")
        lines.append(p(msg))
        if action is not None:
            if callable(action): action = action()
            if isinstance(action, Action):
                url = web.websafe(action.url)
                title = web.websafe(action.title or u"Cliquez ici pour continuer")
                action = u"""<p><a href="%s" accesskey="r">%s</a></p>""" % (url, title)
            lines.append(p(action))
        if exc_info is not None:
            lines.append(u"""<div class="small" style="margin-top: 3em; color: gray;">Pour information, le message d'erreur technique est&nbsp;&nbsp;&nbsp;""")
            lines.extend(traceback.format_exception_only(*exc_info[:2]))
            if showtb:
                lines.append(u"<!--")
                lines.append("".join(traceback.format_exception(*exc_info)))
                lines.append(u"-->")
            lines.append(u"""</div>""")
        lines.append(u"""</div>""")
        return u"\n".join([u(line) for line in lines])
    def __unicode__(self): return self.render()

def set_alert(template=None, action=Undef, delay=None, **params):
    """Décorateur qui permet de gérer automatiquement une instance de Alert dans la
    page.

    * Une instance vide self.alert = Alert() est créée dans la page
    * Si la valeur de retour n'est pas une instance de Alert, aucun traitement
      particulier supplémentaire n'est effectué
    * Si une exception se produit et qu'on n'est pas en mode debug, créer une
      nouvelle instance de Alert avec le message de l'exception.
    * Sinon:
      * Si template==None, afficher une page d'erreur générique avec une alerte
        de type Error.
      * Sinon, mettre à jour la variable alert avec la valeur retournée et
        effectuer le rendu du template spécifié.
    """
    def decorator(method):
        def wrapper(self, *args, **kw):
            self.alert = Alert(action=action)
            try:
                alert = method(self, *args, **kw)
            except:
                if web.config.debug: raise
                else: alert = Alert("Une erreur inattendue s'est produite")
            if isinstance(alert, Alert):
                if alert.type == 'success':
                    title = u"Succès"
                elif alert.type == 'warning':
                    title = u"Avertissement"
                else:
                    title = u"Erreur"
                if action is not Undef and alert.action is None:
                    alert.action = action
                if template is not None:
                    self.alert = alert
                    return self.r(seqof(template), self)
                else:
                    lines = []
                    lines.append(u"""\
<!DOCTYPE html><!-- -*- coding: utf-8 mode: html -*- vim:sw=4:sts=4:et:ai:si:sta:fenc=utf-8
--><html lang="fr">
<head>""")
                    lines.append(load())
                    if alert.action and delay is not None:
                        url = web.websafe(alert.action.url)
                        lines.append(u"""\
<meta http-equiv="Refresh" content="%s; URL=%s"/>""" % (delay, url))
                    lines.append(u"""\
<title>%(title)s</title>
</head>
<body>""" % locals())
                    if web.config.have_session:
                        session = web.config._session
                        if 'menu' in session: lines.append(session.menu.render())
                    lines.append(u"""\
<div class="container">
<h1>%(title)s</h1>""" % locals())
                    lines.append(alert.render())
                    lines.append(u"""\
</div>
</body>
</html>""")
                    return u"\n".join(lines)
            else:
                return alert
        return wrapper
    return decorator
